/*****************************************************************************
 *
 *  Optional.h                                  XMM Science Analysis Subsystem
 *
 *  Template class for optional values which may be 'undefined'
 *
 *  (c) 1998  ESA  -  Written by Jon Brumfitt, Aurora
 *
 *  $Id: Optional.h,v 1.5 2008/09/26 16:38:55 eojero Exp $
 *
 *****************************************************************************/

#ifndef OPTIONAL_H
#define OPTIONAL_H

#include <iostream>
#include <string>


//================================================================
//  Class Optional
//
//  Optional is a template class which allows values to have an
//  extra 'undefined' state. An conversion operator provides implicit
//  conversion to the underlying type, so that the optional value can
//  be used in the same way as the underlying type; however, an explicit
//  cast, or call to value(), may sometimes be needed.
//
//  Optional values are initially undefined, unless they are initialised
//  by the constructor. The object may be defined by assigning a value.
//
//  Examples:
//    Optional<int> x;                      // Initially undefined
//    std::cout << x.isDefined();                // 0
//    x = 42;                               // Defined by assignment
//    std::cout << x << ' ' << x.isDefined();    // 42 1
//    int x1 = x;                           // Implicit conversion
//    int x2 = x.value();                   // Explicit conversion
//
//    Optional<int> y(33);                  // Defined by construction
//    std::cout << y << ' ' << y.isDefined();    // 33 1
//
//    Optional<int> z(y);                   // Defined if 'y' is defined
//
//================================================================

template<class T>
class Optional
{
public:


    // Constructors
    Optional() : _defined(false) {}
    Optional(const T& value) 
      : _value(value), _defined(true) {}
    Optional(const Optional<T>& rhs) 
      : _value(rhs._value), _defined(rhs._defined) {}

    // Assignment operator
    Optional<T>& operator=(const Optional<T>&);
    Optional<T>& operator=(const T&);

    // Test whether value is assigned
    bool isDefined() const {return _defined;}

    // Return the value
    const T& value() const {return _value;}

    // Conversion to underlying type
    operator T() const {return _value;}

private:

    T    _value;
    bool _defined;
};


/*** Assignment operator ***/

template<class T>
Optional<T>& Optional<T>::operator=(const Optional<T>& rhs)
{
    _value   = rhs._value;
    _defined = rhs._defined;

    return *this;
}

template<class T>
Optional<T>& Optional<T>::operator=(const T& rhs)
{
    _value   = rhs;
    _defined = true;

    return *this;
}


/*** Write to stream ***/

template<class T>
std::ostream& operator<<(std::ostream& os, const Optional<T>& p)
{
    return os << p.value();
}

#endif
