// f90cxx.h
// 
// Handling of mixing Fortran and C++
//
// Original by Marco Beijersbergen
//
// Modifications for 64 bit kernels by Eduardo Ojero Pascual
// ESAC 24 Nov 2010
//
// Copyright (c) 1997-2011 European Space Agency

#ifndef f90cxx_h
#define f90cxx_h

#include <vector>
#include <iostream>
#include <cassert>

// Temporarily force this compiler combination

#define NAGF90 
#define GNUCC

#ifdef NAGF90

extern "C" void f90_io_init();
extern "C" void f90_io_finish();

#define f90init f90_io_init
#define f90finish f90_io_finish
 

typedef char Integer1;
typedef short Integer2;
typedef int Integer3;
typedef int Logical;
typedef double Double;
typedef float Single;
typedef char* Pointer;


// On 64 bit kernel the components of 
// Triplet (extent, mult, lower) and the offset
// variable must be all 64 bit integers
//
// We define here new types dope_Offset and dope_Chrlen
// which depend on processor type through
// OBJECT_SIZE_64 and CHRLEN_SIZE_64 preprocessor.
// Then assign these types to extent, mult, lower and offset 
//
// Adapted from NAG dope.h (nagfor 5.2)

#ifdef OBJECT_SIZE_64
typedef long long dope_Offset;
#else
typedef int dope_Offset;
#endif

#if defined(CHRLEN_SIZE_64) && defined(OBJECT_SIZE_64)
typedef long long dope_Chrlen;
#else
typedef int dope_Chrlen;
#endif


struct Triplet {
  dope_Offset extent,mult,lower;

};

template<class T, int N>
struct Dope {

  Dope() : addr(0) {}

  // old Dope
  Dope(const T* vec, const unsigned long* ext, const unsigned long* low=0) {
    addr = const_cast<T*>(vec);		// address to first element

    offset = 0;


    // Swap C++ and Fortran indices
    for (int i = N-1, j = 0; i >= 0; --i, ++j) {
      dim[j].extent	= ext[i];
      dim[j].lower	= (low ? low[i] : 0);

      if (j == 0)
        dim[j].mult = sizeof(T);
      else
        dim[j].mult = dim[j-1].mult * dim[j-1].extent;

      offset -= dim[j].mult * dim[j].lower;

    }
  }

  Dope(const T* vec, const std::vector<unsigned long>& ext,
	const std::vector<unsigned long>& low=std::vector<unsigned long>( 0) ) {
    assert(ext.size() == N);
    assert(low.size() == 0 || low.size() == N);

    addr = const_cast<T*>(vec);		// address to first element

    offset = 0;

    // Swap C++ and Fortran indices
    for (int i = N-1, j = 0; i >= 0; --i, ++j) {
      dim[j].extent	= ext[i];
      dim[j].lower	= (low.size() > 0 ? low[i] : 0);

      if (j == 0)
        dim[j].mult = sizeof(T);
      else
        dim[j].mult = dim[j-1].mult * dim[j-1].extent;

      offset -= dim[j].mult * dim[j].lower;

    }
  }
  //old Dope

  Dope(const T* vec, const uint32_t* ext, const uint32_t* low=0) {
    addr = const_cast<T*>(vec);		// address to first element

    offset = 0;
    
    
    // Swap C++ and Fortran indices
    for (int i = N-1, j = 0; i >= 0; --i, ++j) {
      dim[j].extent	= ext[i];
      dim[j].lower	= (low ? low[i] : 0);

      if (j == 0)
        dim[j].mult = sizeof(T);
      else
        dim[j].mult = dim[j-1].mult * dim[j-1].extent;

      offset -= dim[j].mult * dim[j].lower;
      
    }
  }

  Dope(const T* vec, const std::vector<uint32_t>& ext,
	const std::vector<uint32_t>& low=std::vector<uint32_t>( 0) ) {
    assert(ext.size() == N);
    assert(low.size() == 0 || low.size() == N);

    addr = const_cast<T*>(vec);		// address to first element

    offset = 0;

    // Swap C++ and Fortran indices
    for (int i = N-1, j = 0; i >= 0; --i, ++j) {
      dim[j].extent	= ext[i];
      dim[j].lower	= (low.size() > 0 ? low[i] : 0);

      if (j == 0)
        dim[j].mult = sizeof(T);
      else
        dim[j].mult = dim[j-1].mult * dim[j-1].extent;

      offset -= dim[j].mult * dim[j].lower;

    }
  }


  T* addr;
  dope_Offset offset;
  Triplet dim[N];
};

template<class T, int N>
inline std::ostream& operator<<(std::ostream& os, const Dope<T,N>& d)
{
  os <<"addr="<<(void*)d.addr << " offset="<<d.offset;
  for (int i = 0; i < N; ++i)
    os << " extent="<<d.dim[i].extent << " mult="<<d.dim[i].mult << " lower="<<d.dim[i].lower;
  os << std::endl;
  return os;
}

template<int N>
struct ChDope {

  ChDope() : addr(0) {}
  char* addr;
  dope_Chrlen len;
  dope_Offset offset;
  Triplet dim[N];
};

template<int N>
inline std::ostream& operator<<(std::ostream& os, const ChDope<N>& d)
{
  os <<"addr="<<(void*)d.addr << " offset="<<d.offset;
  for (int i = 0; i < N; ++i)
    os << " extent="<<d.dim[i].extent << " mult="<<d.dim[i].mult << " lower="<<d.dim[i].lower;
  os << std::endl;
  return os;
}

#ifdef SUNCC
#define F90PROCDEF(ret,proc) extern "C" ret proc##_
#define F90MODPROCDEF(ret,mod,proc) extern "C" ret mod##_MP_##proc 
#define F90MODPROCPROTO(ret,mod,proc) extern "C" ret mod##_MP_##proc
#define F90MODPROC(mod,proc) mod##_MP_##proc
#define F90MODVARDEF(type,mod,var) type mod##_MP_##var
#define F90MODVAR(mod,var) mod##_MP_##var
#endif

#ifdef GNUCC
#define F90PROCDEF(ret,proc) extern "C" ret proc##_
#define F90MODPROCDEF(ret,mod,proc) extern "C" ret mod##_MP_##proc
#define F90MODPROCPROTO(ret,mod,proc) extern "C" ret mod##_MP_##proc
#define F90MODPROC(mod,proc) mod##_MP_##proc
#define F90MODVARDEF(type,mod,var) type mod##_MP_##var
#define F90MODVAR(mod,var) mod##_MP_##var
#endif

#endif

#ifdef SUNF90

#define f90init()
#define f90finish()

    struct DopeVector
    {
        void *ptr;                      /* ptr is a pointer to the first
                                         * byte of the array.
                                        */
        unsigned el_len;                /* el_len is the length of an
                                         * element of the array.
                                         * If the type of the array is
                                         * CHARACTER, then the size is
                                         * in bytes; otherwise, the size
                                         * is in bits.
                                        */
        unsigned assoc : 1;             /* assoc is 1 if the dope vector
                                         * is associated.  This field is
                                         * irrelevant for assumed-shape
                                         * arrays, but the same data
                                         * structure is used for
                                         * Fortran 90 pointers.
                                        */
        unsigned ptr_alloc : 1;         /* ptr_alloc is 1 if the array
                                         * referenced by the dope vector
                                         * was allocated using a pointer
                                         * in an ALLOCATE statement.
                                         * This field is irrelevant
                                         * for assumed-shape arrays.
                                        */
        unsigned p_or_a : 2;            /* p_or_a is 1 if the dope vector
                                         * implements a Fortran 90 pointer
                                         * and 2 if it implements an
                                         * allocatable array.  This field
                                         * is irrelevant for assumed-shape
                                         * arrays.
                                        */
        unsigned unused1 : 25;          /* unused1 is not used. */
        unsigned n_dim : 3;             /* n_dim is the rank of the array. */

/* -----  word break  ----- */

        unsigned type : 8;              /* type is the type of the array
                                         *
                                         *    1 == typeless
                                         *    2 == integer
                                         *    3 == real
                                         *    4 == complex
                                         *    5 == logical
                                         *    6 == character
                                         *    7 == character derived type
                                         *    8 == noncharacter derived type
                                        */
        unsigned dpflag : 1;            /* dpflag is 1 if the type was
                                         * double precision or double
                                         * complex
                                        */
        unsigned kind_or_star : 3;      /* kind_or_star is 1 if a KIND
                                         * parameter appeared in the
                                         * declaration of the array.
                                         * kind_or_star is 2 if type
                                         * was star sized (e.g. INTEGER*4)
                                        */
        unsigned int_len : 12;          /* int_len is the length of an
                                         * element in bits.
                                        */
        unsigned dec_len : 8;           /* dec_len is the length of an
                                         * element in bytes if
                                         * kind_or_star is nonzero.
                                        */

        void *orig_base;                /* orig_base is a pointer to
                                         * the allocated object containing
                                         * the array referenced by the
                                         * dope vector
                                        */
        unsigned orig_size;             /* orig_size is the size in bytes of
                                         * the allocated object containing
                                         * the array referenced by the
                                         * dope vector.
                                        */
    };

    struct DvDimen
    {
        int low_bound;          /* low_bound is the lower bound in the
                                   i-th dimension.
                                */
        int extent;             /* extent is the extent in the i-th
                                 * dimension.  Extent is one greater
                                 * than the difference between the
                                 * upper and lower bounds.  The upper
                                 * bound can be computed from the
                                 * lower bound and the extent.
                                */
        int stride_mult;        /* stride_mult is distance in elements
                                 * between consecutive elements in the
                                 * i-th dimension.  This field is needed
                                 * for discontiguous arrays.
                                */
    };

template<int N> struct Dope : public DopeVector {
  DvDimen dim[N];
};

struct Dope1 : public DopeVector {
  DvDimen dim[1];
};

struct DopeN : public DopeVector {
  DvDimen dim[10];
};

enum F90Type { F90Typeless = 1, F90Integer, F90Real, F90Complex, F90Logical, F90Char, F90CharDerived, F90NonCharDerived };

inline std::ostream& operator<<(std::ostream& os, const DopeN& d)
{
  os <<"ptr="<<d.ptr << " el_len="<<d.el_len << " assoc="<<d.assoc
     <<" ptr_alloc="<<d.ptr_alloc << " p_or_a="<<d.p_or_a 
     <<" unused1="<<d.unused1 << " n_dim="<<d.n_dim << " type="<<d.type 
     << " dpflag="<<d.dpflag << " kind_or_star="<<d.kind_or_star
     <<" int_len="<<d.int_len << " dec_len="<<d.dec_len
     <<" orig_base="<<d.orig_base << " orig_size="<<d.orig_size;
  for (int i = 0; i < d.n_dim; ++i)
    os << " low_bound="<<d.dim[i].low_bound << " extent="<<d.dim[i].extent << " stride_mult="<<d.dim[i].stride_mult;
  os << std::endl; 

  unsigned elements = 1;
  for (int ie = 0; ie < d.n_dim; ++ie)
    elements *= d.dim[ie].extent;
  for (int ir = 0; ir < elements; ++ir) {
    switch (d.type) {
      case F90Integer: os << ((int*)d.ptr)[ir] << ' '; break;
      case F90Real: os << (d.dpflag==1?((double*)d.ptr)[ir] : ((float*)d.ptr)[ir])<< ' '; break;
      default: os << " invalid type" << std::endl;
    }
  }
  os << std::endl;
  return os;
}

inline std::ostream& operator<<(std::ostream& os, const Dope<1>& d)
{ return os << *(const DopeN*)&d; }

inline std::ostream& operator<<(std::ostream& os, const Dope<2>& d)
{ return os << *(const DopeN*)&d; }

#ifdef SUNCC
#define F90PROCDEF(ret,proc) extern "C" ret proc##_
#define F90MODPROCPROTO(ret,mod,proc) extern "C" ret mod##_##proc
#define F90MODPROCDEF(ret,mod,proc) extern "C" ret mod##_##proc
#define F90MODPROC(mod,proc) mod##_##proc 
#define F90MODVARDEF(type,mod,var) type mod##_##var
#define F90MODVAR(mod,var) mod##_##var
#endif

#ifdef GNUCC
#define F90PROCDEF(ret,proc) extern "C" ret proc##_
#define F90MODPROCPROTO(ret,mod,proc) extern "C" ret mod##_##proc () asm(#proc##".in."#mod##"_"); ret mod##_##proc
#define F90MODPROCDEF(ret,mod,proc) extern "C" ret mod##_##proc () asm(#proc##".in."#mod##"_"); ret mod##_##proc
#define F90MODPROC(mod,proc) mod##_##proc
#define F90MODVARDEF(type,mod,var) type mod##_##var
#define F90MODVAR(mod,var) mod##_##var
#endif

#endif

// Backwards compatibility for CAL/OAL

typedef Dope<char,1> Dope1;
typedef Dope<char,2> Dope2;
typedef Dope<char,3> Dope3;
typedef Dope<char,4> Dope4;
typedef Dope<char,5> Dope5;

#endif
