!-----------------------------------------------------------------------------
!
!                       XMM Science Analysis System (SAS)
!                      (c) 1997 - 2010 European Space Agency
!
!-----------------------------------------------------------------------------
!
!   random.f90 --- A good random generator ---
!
!   DESCRIPTION:
!       The module provides access from Fortran-90 to the single global
!       random number generator that is part of the SAS infrastructure.
!       The generator is known to be one of the best freely available
!       ones. It possesses a period which can be considered as 'Infinity'
!       for all practical purposes and passes all crucial statistical
!       tests. The code is a "Subtract-and-borrow" random number generator
!       proposed by	Marsaglia and Zaman, implemented by F. James with the name
!		RCARRY in 1991, and later improved by Martin Luescher
!		in 1993 to produce "Luxury Pseudorandom Numbers".
!		Fortran 77 coded by F. James, 1993
!       For more information consult the source files 'Ranlux.h'/'Random.h'
!
!   REFERENCES:
!		M. Luscher, Computer Physics Communications  79 (1994) 100
!		F. James, Computer Physics Communications 79 (1994) 111
!
!   USAGE EXAMPLES:
!       program randtest
!           use Random
!           implicit none
!           integer i
!
!           call SAS_randSetMinMaxRange(0., 1., CLOSED)
!           write(*, *) 'Seed is: ', SAS_randSeed()
!           do i=1, 100
!               write(*, *) SAS_rand()
!           enddo
!       end program
!
!   CHANGE HISTORY:
!       UL, 2002-11-05: 1.0.0: - first version (adapted from Ranlux.h)
!
!   AUTHOR:
!       U. Lammers, Tue Nov  5 15:22:24 MET 2002
!
!   REFERENCES:
!       [1]: Software Specification Document for the XMM Science Analysis
!            Subsystem, M. Dahlem et. al.,
!
!   $Header: /sasbuild/repositories/sasdev/sas/packages/utils/src/random.f90,v 1.2 2008/09/26 16:38:55 eojero Exp $
!
module Random
    use types
    implicit none

    !
    !   range types; symbolic constants specifying the kind of interval
    !   of pseudo-random numbers returned by SAS_rand() below
    !
    integer(kind=int32), parameter, public :: OPEN = 0, &
                                              OPEN_CLOSED = 1, &
                                              CLOSED_OPEN = 2, &
                                              CLOSED = 3

    interface

    function SAS_randRangeType() result(rangeType)
        use types
        integer(kind=int32)                      :: rangeType
    end function
        !
        !   PURPOSE:
        !       Return interval range type currently active in generator
        !

       
    function SAS_randSeed() result(seed)
        use types
        integer(kind=int32)                      :: seed
    end function
        !
        !   PURPOSE:
        !       Return seed that generator was initialized/instantiated with;
        !


    function SAS_randMin() result(min)
        use types
        real(kind=double)                        :: min
    end function

    function SAS_randMax() result(max)
        use types
        real(kind=double)                        :: max
    end function
        !
        !   PURPOSE:
        !       Return lower/upper bound of numbers returned by SAS_rand()
        !       (see below)
        !


    subroutine SAS_randSetMin(min)
        use types
        real(kind=double), intent(in)            :: min
    end subroutine

    subroutine SAS_randSetMax(max)
        use types
        real(kind=double), intent(in)            :: max
    end subroutine

    subroutine SAS_randSetMinMaxRange(min, max, rangeType)
        use types
        real(kind=double), intent(in)            :: min, max
        integer(kind=int32), intent(in)          :: rangeType
    end subroutine
        !
        !   PURPOSE:
        !       Set lower/upper bound + range type of numbers returned by
        !       subsequent invocations of SAS_rand() below
        !
        !   ARGUMENTS:
        !       min/max  : lower/upper bound of pseudo-random number interval
        !       rangeType: type of random number interval (see above)
        !


    subroutine SAS_randPopState()
    end subroutine
        !
        !   PURPOSE:
        !       Pop state of random number generator from a stack; popping
        !       if stack is empty has no effect; see also below
        !       SAS_randPushState()
        !


    subroutine SAS_randReset()
    end subroutine
        !
        !   PURPOSE:
        !       Reset the generator to the state it was in right after
        !       construction
        !


    function SAS_rand() result(ran)
        use types
        real(kind=double)                        :: ran
    end function
        !
        !   PURPOSE:
        !       Return a uniformily distributed pseudo-random number in the
        !       range
        !           o (min, max)
        !           o (min, max]
        !           o [min, max)
		!           o [min, max]
        !       depending on the value of <rangeType> specified upon
        !       object construction or set via the dedicated routine
        !       SAS_randSetMinMaxRange() above
        !

    end interface

    interface SAS_randPushState
        subroutine SAS_randPushState1()
        end subroutine

        subroutine SAS_randPushState2(min, max, rangeType)
            use types
            real(kind=double), intent(in)            :: min, max
            integer(kind=int32), intent(in)          :: rangeType
        end subroutine
            !
            !   PURPOSE:
            !       Push state of random number generator onto a stack;
            !       second form is equivalent to calling sequence
            !           call SAS_randPushState()
            !           call SAS_randSetMinMaxType(min, max, rangeType)
            !
    end interface

end module Random
