!-----------------------------------------------------------------------------
!
!                       XMM Science Analysis System (SAS)
!                      (c) 1997 - 2010 European Space Agency
!
!-----------------------------------------------------------------------------
!
!   stime.f90 --- arithmetic with times ---
!
!   DESCRIPTION:
!       Instances of type STimeHandleT (from stimetypes.f90) represent valid
!       points in time. Constructors and associated functions/subroutines
!       allow to convert between different representations, including
!       Julian/Modified Julian day numbers and calendar day/time.
!       Basic arithmetics is available in the form of functions
!       addTo/subtractFrom/add/diff which take floating point values as
!       operands. These are interpreted as time offsets in seconds.
!       The subraction of two time objects t1 and t2 yields
!       the time difference in seconds between t1 and t2.
!       Note: The module Xmm (declard in xmm.f90 of packages caloalutils)
!       defines a mission-wide reference time which is a static STimeHandleT
!       instance.
!
!       Manipulation of time instances are guarenteed to maintain
!       precision to the micro-sec level. Leap seconds are properly
!       taken into account.
!
!       There are various constructors to construct a time object from either
!       a date/time string, a year/month/day/hour/min/secs tuple or
!       a JD/MJD number. If the construction fails due to format errors
!       or out-of-range conditions a fatal error condition is raised.
!       Alternatively the success of the operation can be checked
!       a-priori with corresponding logical isValidTime() functions
!       (see below).
!
!   USAGE EXAMPLES:
!       program xmmf90test
!       use Xmm
!       use STime
!       implicit none
!
!       type(STimeHandleT)     :: t1, t2, tm, t3
!       real(kind=double)      :: t
!
!       t1 = ST_new('1997-10-23T14:15:34.123')          ! define time t1
!       t2 = ST_add(t1, 10._double)                     ! t1 10 seconds later
!       tm = ST_add(XMM_missionReferenceTime(), 156.1_double)
!                                                       ! 156 seconds after ref
!                                                       !  time
!       t3 = ST_clone(t2)                               ! copy t2 to t3
!       call ST_subtractFrom(t3, 10. * ST_secondsPerDay()) ! t3 10 days earlier
!
!       write(*, *) 'difference between dates ', &
!           trim(ST_toString(t2)), ' and ', trim(ST_toString(t3))
!       write(*, *) 'is: ', ST_diff(t2, t3), 'seconds.'
!
!       write(*, *) 'Modified Julian day number corresponding to ', &
!           trim(ST_toString(t3)), ' is ', &
!           ST_julianDayFull(ST_modifiedJulianDay(t3))
! 
!       write(*, *) 'Mission reference STime is: ', &
!           trim(ST_toString(XMM_missionReferenceTime()))
!
!       write(*, *) '10.3 seconds later is ', &
!           trim(ST_toString(ST_add(XMM_missionReferenceTime(), 10.3_double)))
!
!       write(*, *) 'XMM mission start time is    : ', &
!           trim(ST_toString(XMM_missionStartTime()))
!       write(*, *) 'XMM mission end time is      : ', &
!           trim(ST_toString(XMM_missionEndTime()))
!
!       write(*, *) 'The current time is ', ST_toString(ST_currentTime())
!
!       t = 110817735.184
!       write(*, *) 'MET tag ', t, '= ', &
!           trim(ST_toString(ST_add(XMM_missionReferenceTime(), t)))
!
!       write(*, *) 'All tests sucessfully passed'
!    end program
!
!   CHANGE HISTORY:
!       UL, 2001-07-12: 1.0.0: - first version (adapted from STime.h)
!       UL, 2001-07-13: 1.0.1: - added forgotten optional parameter to
!                                ST_{modified}JulianDay()
!
!   AUTHOR:
!       U. Lammers, Thu Jul 12 15:00:17 MET DST 2001
!
!   REFERENCES:
!       [1]: Software Specification Document for the XMM Science Analysis
!            Subsystem, M. Dahlem et. al.,
!
!   $Header: /sasbuild/repositories/sasdev/sas/packages/utils/src/stime.f90,v 1.5 2008/09/26 16:38:55 eojero Exp $
!
module STime
    use types
    use STimeTypes
    implicit none

    !
    !   construct new STime instances
    !
    interface ST_new

        function ST_new0() result(time)
            use stimetypes
            type(STimeHandleT)                        :: time
        end function
            !
            !   PURPOSE:
            !       Instantiate a STime object set to ST_earliest() [see below]
            !       i.e., ST_isSameInstant(t, ST_earliest())
            !       yields true
            !


        function ST_new1(timeString) result(time)
            use stimetypes
            character(len=*), intent(in)              :: timeString
            type(STimeHandleT)                        :: time
        end function
            !
            !   PURPOSE:
            !       Instantiate a STime object from a given time specified
            !       as a string adhereing to either of the formats
            !       FITS/oldFITS/CALENDAR:
            !       - yyyy-mm-dd[Thh:mm:ss[.msec]]  "1997-08-12T23:12:45.1233"
            !       - dd/mm/yy                      "31/12/99"
            !       - Wkd Mon day hh:mm::ss yyyy    "Sun Sep 16 01:03:52 1973"
            !       An error condition occurs if the passed time string is
            !       invalid or one or more numeric fields are out of range
            !       (see above for allowd limits). In this the object will
            !       be set to ST_earliest()
            !       The construct ST_new('now') yields an STime object
            !       corresponding to the current time.
            !
            !  ARGUMENTS:
            !       I   : dateTimeString: time string in format according to
            !                             <format> parameter or "now" or "NOW"
            !                             in the latter to cases the STime
            !                             object represents the current time
            !                             (alias for ST_currentTime() - see
            !                             below)
            !                             please note: the evolution around
            !                             the introduction of a leap second is:
            !                               1997-06-30T23:59:59.3   - before
            !                               1997-06-30T23:59:60.3   - during
            !                               1997-07-01T00:00:00.3   - after
            !                             at this point in time TAI is 31s
            !                             ahead of UTC;
            !


        function ST_new2(year, month, day, hour, minute, secs, &
                         excessSecs) result(time)
            use stimetypes
            integer(kind=int32), intent(in)           :: year
            integer(kind=int32), intent(in), optional :: month, day, hour, &
                                                         minute, secs
                                                         ! d/f: 1/1/0/0/0
            real(kind=double), intent(in), optional   :: excessSecs ! d/f: 0.
            type(STimeHandleT)                        :: time
        end function
            !
            !   PURPOSE:
            !       Instantiate a STime object from a
            !       {year/month/day/hour/minute/secs/excessSecs} tuple
            !       which represents a point in time expressed in UTC.
            !       Passing out-of-range arguments will result in an
            !       error condition. For checking if STime construction from
            !       a {year/month/day/hour/minute/secs/excessSecs} tuple will
            !       succeed, use ST_isValidTime() below.
            !
            !  ARGUMENTS:
            !      I   : year           : year number [1958, 2029]
            !            month          : month number [1, 12]
            !            day            : day number [1, 31]
            !            hour           : hours since midnight [0, 23]
            !            minute         : minutes since hour [0, 59]
            !            secs           : seconds since minute [0, 60] (allow
            !                             one leap second)
            !            excessSecs : value in seconds to be added to time -
            !                             normally (not necessarily)
            !                             0 <= excessSecs < 1
            !
            !       Please note: The specified boundaries for the various
            !                    fields are nominal ones but actually given
            !                    values can exceed these bounds. In case they
            !                    do, the number are interpreted as offsets from
            !                    the time instance
            !                        year-01-01T00:00:00
            !                    Example: 1980-14-45T100:111:222 corresponds to
            !                    1981-03-21T05:54:43.000, which is 222 seconds,
            !                    111 minutes, 100 hours, 45 (Julian) days,
            !                    14 months after 1980-01-01T00:00:00.
            !


        function ST_new3(timeVal, r) result(time)
            use stimetypes
            real(kind=double), intent(in)             :: timeVal
            integer(kind=int32), intent(in), optional :: r  ! d/f: ST_MJD
            type(STimeHandleT)                        :: time
        end function
            !
            !   PURPOSE:
            !       Instantiate an STime object from double value representing
            !       a Julian or Modified Julian day number in UT or a decimal
            !       year number (e.g. 1998.12345). Passing negative values
            !       will result in an error condition.
            !
            !  ARGUMENTS:
            !       I   : timeVal   : time value corresponding to <t>; has to
            !                         be positive
            !             r         : switch determining representation of
            !                         <timeVal> (see stimetypes.f90 for list
            !                         of allowed values)
            !


        function ST_new4(timeValInt, timeValFrac, r) result(time)
            use stimetypes
            integer(kind=int32), intent(in)           :: timeValInt
            real(kind=double), intent(in)             :: timeValFrac
            integer(kind=int32), intent(in), optional :: r  ! d/f: MJD
            type(STimeHandleT)                        :: time
        end function
            !
            !   PURPOSE:
            !       Instantiate a STime object from an integer/double pair
            !       representing a time as a Julian or Modified Julian day
            !       number broken down into integer and fractional part in UT.
            !
            !  ARGUMENTS:
            !       timeValInt  : integer part of time value
            !       timeValFrac : fractional part of time value; has to be
            !                     positive
            !       r           : switch determining representation of
            !                     <timeVal> (see stimetypes.f90 for list
            !                     of allowed values
            !
    end interface ST_new


    !
    !   check validity of times/if construction of STimes with above
    !   constructors will succeed; arguments are identical to corresponding
    !   routines above
    !
    interface ST_isValidTime

        logical function ST_isValidTime1(timeString)
            use stimetypes
            character(len=*), intent(in)              :: timeString
        end function

        logical function ST_isValidTime2(year, month, day, hour, minute, &
                                         secs, excessSecs)
            use stimetypes
            integer(kind=int32), intent(in)           :: year
            integer(kind=int32), intent(in), optional :: month, day, hour, &
                                                         minute, secs
                                                         ! d/f: 1/1/0/0/0
            real(kind=double), intent(in), optional   :: excessSecs
                                                         ! d/f: 0.
        end function

        logical function ST_isValidTime3(timeVal, r)
            use stimetypes
            real(kind=double), intent(in)             :: timeVal
            integer(kind=int32), intent(in), optional :: r  ! d/f: ST_MJD
        end function

    end interface ST_isValidTime

    interface ST_release
        !
        !   PURPOSE:
        !       Release STime objects obtained with either of above constructor
        !       routines.
        !
        subroutine ST_releaseTime(time)
            use stimetypes
            type(STimeHandleT), intent(in)            :: time
        end subroutine

        subroutine ST_releaseJulianDay(julianDay)
            use stimetypes
            type(JulianDayHandleT), intent(in)      :: julianDay
        end subroutine

    end interface ST_release

    interface

    !
    !   auxiliary routines
    !
    function ST_clone(timerhs) result(time)
        use stimetypes
        type(STimeHandleT), intent(in)            :: timerhs
        type(STimeHandleT)                        :: time
    end function
        !
        !   PURPOSE:
        !       copy constructor - initialize object from reference to
        !       another one
        !


    subroutine ST_assign(timelhs, timerhs)
        use stimetypes
        type(STimeHandleT), intent(in)            :: timelhs, timerhs
    end subroutine
        !
        !   PURPOSE:
        !       Assign existing STime to another existing STime
        !


    !
    !   inquiry methods
    !

    !   year number [1958, 2029]
    function ST_year(time) result(year)
        use stimetypes
        type(STimeHandleT), intent(in)            :: time
        integer(kind=int16)                       :: year
    end function

    !   month number [1, 12]
    function ST_month(time) result(month)
        use stimetypes
        type(STimeHandleT), intent(in)            :: time
        integer(kind=int16)                       :: month
    end function

    !  day number [1, 31] 
    function ST_day(time) result(day)
        use stimetypes
        type(STimeHandleT), intent(in)            :: time
        integer(kind=int16)                       :: day
    end function

    !   hours elapsed since midnight [0, 23]
    function ST_hour(time) result(hour)
        use stimetypes
        type(STimeHandleT), intent(in)            :: time
        integer(kind=int16)                       :: hour
    end function

    !   minutes elapsed since hour [0, 59]
    function ST_minute(time) result(minute)
        use stimetypes
        type(STimeHandleT), intent(in)            :: time
        integer(kind=int16)                       :: minute
    end function

    !   seconds since minute [0, 60] (allow one leap second)
    function ST_seconds(time) result(seconds)
        use stimetypes
        type(STimeHandleT), intent(in)            :: time
        integer(kind=int16)                       :: seconds
    end function

    !   0=Sun, 1=Mon, etc.
    function ST_weekday(time) result(weekday)
        use stimetypes
        type(STimeHandleT), intent(in)            :: time
        integer(kind=int8)                        :: weekday
    end function

    !   0 <= fracSeconds <= 1
    function ST_fracSeconds(time) result(fracSeconds)
        use stimetypes
        type(STimeHandleT), intent(in)            :: time
        real(kind=double)                         :: fracSeconds
    end function

    !   number of leap seconds
    function ST_leapSeconds(time) result(leapSeconds)
        use stimetypes
        type(STimeHandleT), intent(in)            :: time
        integer(kind=int16)                       :: leapSeconds
    end function


    !
    !   arithmetic
    !

    !   add <delta> seconds to <time>
    subroutine ST_addTo(time, delta)
        use stimetypes
        type(STimeHandleT), intent(in)            :: time
        real(kind=double), intent(in)             :: delta
    end subroutine


    !   subtract <delta> seconds from <time>
    subroutine ST_subtractFrom(time, delta)
        use stimetypes
        type(STimeHandleT), intent(in)            :: time
        real(kind=double), intent(in)             :: delta
    end subroutine


    !   add <seconds> seconds to time returning new STime object
    function ST_add(time, seconds) result(newTime)
        use stimetypes
        type(STimeHandleT), intent(in)          :: time
        real(kind=double), intent(in)           :: seconds
        type(STimeHandleT)                      :: newTime
    end function


    !   subtract <seconds> seconds from time returning new STime object
    function ST_sub(time, seconds) result(newTime)
        use stimetypes
        type(STimeHandleT), intent(in)          :: time
        real(kind=double), intent(in)           :: seconds
        type(STimeHandleT)                      :: newTime
    end function


    !   compute time difference between <t0> and <t1> in seconds
    function ST_diff(t0, t1) result(difftime)
        use stimetypes
        type(STimeHandleT), intent(in)          :: t0, t1
        real(kind=double)                       :: difftime ! t0-t1 [s]
    end function


    !
    !   miscalleneous functions
    !

    !   return current-time object as GMT (asGMT=.true.) or local time
    !   (asGMT=.false.)
    function ST_currentTime(asGMT) result(time)
        use stimetypes
        Logical, intent(in), optional            :: asGMT ! d/f: .true.
        type(STimeHandleT)                       :: time
    end function


    !
    !   Construct JuliandDayNumber object from integer and fractional day
    !   number
    function ST_julianDayNew(day, fractionalDay) result(julianDay)
        use stimetypes
        integer(kind=int32), intent(in)         :: day
        real(kind=double), intent(in)           :: fractionalDay
        type(JulianDayHandleT)                  :: julianDay
    end function


    !   return integer day number of JulianDay number object
    function ST_julianDayDay(julianDay) result(day)
        use stimetypes
        type(JulianDayHandleT), intent(in)      :: julianDay
        integer(kind=int32)                     :: day
    end function


    !   set integer day of JulianDay number object
    subroutine ST_julianDaySetDay(julianDay, day)
        use stimetypes
        type(JulianDayHandleT), intent(in)      :: julianDay
        integer(kind=int32), intent(in)         :: day
    end subroutine


    !   return fractional day part of JulianDay number object
    function ST_julianDayFracDay(julianDay) result(fracDay)
        use stimetypes
        type(JulianDayHandleT), intent(in)      :: julianDay
        real(kind=double)                       :: fracDay
    end function


    !   set fractional day part of JulianDay number object
    subroutine ST_julianDaySetFracDay(julianDay, fracday)
        use stimetypes
        type(JulianDayHandleT), intent(in)      :: julianDay
        real(kind=double), intent(in)           :: fracDay
    end subroutine


    !   return full Julian day number of Julian day number object as
    !   floating point value
    function ST_julianDayFull(julianDay) result(day)
        use stimetypes
        type(JulianDayHandleT), intent(in)      :: julianDay
        real(kind=double)                       :: day
    end function


    !   construct Julian day number object from STime object
    function ST_julianDay(time, timeSystem) result(julianDay)
        use stimetypes
        type(STimeHandleT), intent(in)          :: time
        integer(kind=int32), intent(in), optional :: timeSystem
        type(JulianDayHandleT)                  :: julianDay
    end function


    !   construct Julian day number object from STime object as MJD
    function ST_modifiedJulianDay(time, timeSystem) result(modifiedJulianDay)
        use stimetypes
        type(STimeHandleT), intent(in)          :: time
        integer(kind=int32), intent(in), optional :: timeSystem
        type(JulianDayHandleT)                  :: modifiedJulianDay
    end function


    function ST_siderealTime(time, longitude) result(sidTime)
        use stimetypes
        type(STimeHandleT), intent(in)          :: time
        real(kind=double), intent(in), optional :: longitude  ! [rad], d/f: 0.
        real(kind=double)                       :: sidTime
    end function
        !
        !   PURPOSE:
        !       Compute the sidereal time at a given longitude on the
        !       surface of the Earth; unit is rad, i.e., 2Pi=24h
        !
        !   ARGUMENTS:
        !       longitude:  longitude of the point on the Earth for which
        !                   sidereal time is to be computed; units: rad
        !                   longitude is 0 on the Greenwich-Meridian,
        !                   positively increasing eastwards
        !


    !   return number of seconds in a minute
    function ST_secondsPerMinute() result(seconds)
        use stimetypes
        real(kind=double)                       :: seconds
    end function

 
    !   return number of seconds in an hour
    function ST_secondsPerHour() result(seconds)
        use stimetypes
        real(kind=double)                       :: seconds
    end function


    !   return number of seconds in a day
    function ST_secondsPerDay() result(seconds)
        use stimetypes
        real(kind=double)                       :: seconds
    end function


    !   return earliest possible STime object
    function ST_earliest() result(time)
        use stimetypes
        type(STimeHandleT)                      :: time
    end function


    !   return latest possible STime object
    function ST_latest() result(time)
        use stimetypes
        type(STimeHandleT)                      :: time
    end function


    !
    !   comparison functions
    !

    !   true if <t0> is later than <t1>
    logical function ST_isLaterThan(t0, t1)
        use stimetypes
        type(STimeHandleT), intent(in)          :: t0, t1
    end function


    !   true if <t0> is later than or same instant as <t1>
    logical function ST_isLaterThanOrSameInstant(t0, t1)
        use stimetypes
        type(STimeHandleT), intent(in)          :: t0, t1
    end function


    !   true if <t0> is earlier than <t1>
    logical function ST_isEarlierThan(t0, t1)
        use stimetypes
        type(STimeHandleT), intent(in)          :: t0, t1
    end function


    !   true if <t0> is earlier than or same instant as <t1>
    logical function ST_isEarlierThanOrSameInstant(t0, t1)
        use stimetypes
        type(STimeHandleT), intent(in)          :: t0, t1
    end function


    !   true of <t0> and <t1> are the same time instant
    logical function ST_isSameInstant(t0, t1)
        use stimetypes
        type(STimeHandleT), intent(in)          :: t0, t1
    end function


    !   true if <t0> and <t1> are at different times
    logical function ST_isNotSameInstant(t0, t1)
        use stimetypes
        type(STimeHandleT), intent(in)          :: t0, t1
    end function

    end interface

    interface ST_toString
        character(len=26) function ST_toFixedString(time, fracSecondDigits, &
                                                    form)
            use stimetypes
            type(STimeHandleT), intent(in)          :: time
            integer, intent(in), optional           :: fracSecondDigits, form
                                                       ! d/f: 3/ST_FITS
        end function
            !
            !  PURPOSE:
            !      Convert time to string representation according to
            !      format specifier (see above); default is FITS format;
            !      <fracSecondDigits> determines the number of digits the
            !      fractional seconds part is printed with;
            !      fracSecondDigits = 0 also suppresses the decimal point.

    end interface

end module STime
